/*
	File                 : NSLTest.h
	Project              : LabPlot
	Description          : NSL tests
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2019 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/
#ifndef NSLTEST_H
#define NSLTEST_H

#include "../CommonTest.h"

class NSLTest : public CommonTest {
	Q_OBJECT
};
#endif
