/*
	File                 : nsl_common.h
	Project              : LabPlot
	Description          : NSL common stuff
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2016 Stefan Gerlach <stefan.gerlach@uni.kn>
	SPDX-License-Identifier: GPL-2.0-or-later
*/

#ifndef NSL_COMMON_H
#define NSL_COMMON_H

#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define i18n(m) m

#endif /* NSL_COMMON_H */
