/*
	File                 : AnalysisTest.h
	Project              : LabPlot
	Description          : Tests any analysis method
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2018 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/
#ifndef ANALYSISTEST_H
#define ANALYSISTEST_H

#include "../CommonTest.h"

class AnalysisTest : public CommonTest {
	Q_OBJECT
};
#endif
