/*
	File                 : NSLTest.cpp
	Project              : LabPlot
	Description          : NSL Tests
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2019 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "NSLTest.h"
#include "backend/core/column/Column.h"
