/*
	File                 : filters.h
	Project              : LabPlot
	Description          : header including all filters
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2018 Stefan-Gerlach <stefan.gerlach@uni.kn>
	SPDX-License-Identifier: GPL-2.0-or-later
*/

#ifndef FILTERS_H
#define FILTERS_H

#include "backend/datasources/filters/AsciiFilter.h"
#include "backend/datasources/filters/BinaryFilter.h"
#include "backend/datasources/filters/ExcelFilter.h"
#include "backend/datasources/filters/FITSFilter.h"
#include "backend/datasources/filters/HDF5Filter.h"
#include "backend/datasources/filters/ImageFilter.h"
#include "backend/datasources/filters/JsonFilter.h"
#include "backend/datasources/filters/MatioFilter.h"
#include "backend/datasources/filters/NetCDFFilter.h"
#include "backend/datasources/filters/ROOTFilter.h"
#include "backend/datasources/filters/ReadStatFilter.h"
#include "backend/datasources/filters/SpiceFilter.h"
#include "backend/datasources/filters/VectorBLFFilter.h"

#endif
