/*
	File                 : TextLabelTest.h
	Project              : LabPlot
	Description          : Tests for TextLabel
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2022 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/

#ifndef TEXTLABELTEST_H
#define TEXTLABELTEST_H

#include "../../CommonTest.h"

class TextLabelTest : public CommonTest {
	Q_OBJECT

private Q_SLOTS:

	void addPlot();
	void multiLabelEditColorChange();
	void multiLabelEditTextChange();
	void multiLabelEditColorChangeSelection();
};

#endif
