/*
	File                 : filter.h
	Project              : AbstractColumn
	Description          : Conversion filter header.
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2017-2020 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/
#ifndef FILTER_H
#define FILTER_H

#include "backend/core/datatypes/BigInt2DateTimeFilter.h"
#include "backend/core/datatypes/BigInt2DayOfWeekFilter.h"
#include "backend/core/datatypes/BigInt2DoubleFilter.h"
#include "backend/core/datatypes/BigInt2IntegerFilter.h"
#include "backend/core/datatypes/BigInt2MonthFilter.h"
#include "backend/core/datatypes/BigInt2StringFilter.h"
#include "backend/core/datatypes/DateTime2BigIntFilter.h"
#include "backend/core/datatypes/DateTime2DoubleFilter.h"
#include "backend/core/datatypes/DateTime2IntegerFilter.h"
#include "backend/core/datatypes/DateTime2StringFilter.h"
#include "backend/core/datatypes/DayOfWeek2BigIntFilter.h"
#include "backend/core/datatypes/DayOfWeek2DoubleFilter.h"
#include "backend/core/datatypes/DayOfWeek2IntegerFilter.h"
#include "backend/core/datatypes/Double2BigIntFilter.h"
#include "backend/core/datatypes/Double2DateTimeFilter.h"
#include "backend/core/datatypes/Double2DayOfWeekFilter.h"
#include "backend/core/datatypes/Double2IntegerFilter.h"
#include "backend/core/datatypes/Double2MonthFilter.h"
#include "backend/core/datatypes/Double2StringFilter.h"
#include "backend/core/datatypes/Integer2BigIntFilter.h"
#include "backend/core/datatypes/Integer2DateTimeFilter.h"
#include "backend/core/datatypes/Integer2DayOfWeekFilter.h"
#include "backend/core/datatypes/Integer2DoubleFilter.h"
#include "backend/core/datatypes/Integer2MonthFilter.h"
#include "backend/core/datatypes/Integer2StringFilter.h"
#include "backend/core/datatypes/Month2BigIntFilter.h"
#include "backend/core/datatypes/Month2DoubleFilter.h"
#include "backend/core/datatypes/Month2IntegerFilter.h"
#include "backend/core/datatypes/SimpleCopyThroughFilter.h"
#include "backend/core/datatypes/String2BigIntFilter.h"
#include "backend/core/datatypes/String2DateTimeFilter.h"
#include "backend/core/datatypes/String2DayOfWeekFilter.h"
#include "backend/core/datatypes/String2DoubleFilter.h"
#include "backend/core/datatypes/String2IntegerFilter.h"
#include "backend/core/datatypes/String2MonthFilter.h"

#endif // ifndef FILTER_H
