/*
	File                 : MatrixFormulaTest.h
	Project              : LabPlot
	Description          : Tests for formula in matrix
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2022 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/

#ifndef MATRIXFORMULATEST_H
#define MATRIXFORMULATEST_H

#include "../CommonTest.h"

class MatrixFormulaTest : public CommonTest {
	Q_OBJECT

private Q_SLOTS:
	void formula1();
	// void formulax();
	// void formulaxpy();
};

#endif
