/*
	File                 : MatrixTest.cpp
	Project              : LabPlot
	Description          : Tests for the Matrix
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2022 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "MatrixTest.h"
#include "backend/matrix/Matrix.h"
#include "commonfrontend/matrix/MatrixView.h"

// TODO

QTEST_MAIN(MatrixTest)
