/*
	SPDX-FileCopyrightText: 2012 Massachusetts Institute of Technology
	SPDX-License-Identifier: MIT
*/

/* The Faddeeva.cc file contains macros to let it compile as C code
   (assuming C99 complex-number support), so just #include it. */
#include "Faddeeva.cc"
